#ifndef SVF_PARSER_H
#define SVF_PARSER_H

#include <vector>
#include <deque>
#include <list>
#include <map>
#include <string>
#include <iosfwd>
#include <boost/shared_ptr.hpp>

#include "mlink/jtag_state.h"

namespace mlink {

    typedef std::vector<bool> JtagBits;

    class JtagStep {
    public:
        JtagStep()
            : tms(0), tdi(0), smask(0), tdo(0), mask(0) { }
        JtagStep(bool _tms, bool _tdi, bool _smask, bool _tdo, bool _mask)
            : tms(_tms), tdi(_tdi), smask(_smask), tdo(_tdo), mask(_mask) { }
        bool tms;
        bool tdi;
        bool smask;
        bool tdo;
        bool mask;
    };

    class SvfShiftOptions;
    class JtagShift: public std::deque<JtagStep> {
    public:
        JtagShift() { }
        explicit JtagShift(const SvfShiftOptions &);
    };

    std::ostream &operator<< (std::ostream & os, const JtagShift &r);

    class SvfOp {
    public:
        virtual ~SvfOp() { }
        //  virtual void print() const { std::cout << "Op: " << name << std::endl; }
        virtual JtagShift getSequence() const { return JtagShift() ; }
        virtual double getDelay() const { return 0; }
        virtual std::string getComment() const { return std::string(); }
        virtual JTAG_STATE beginState() const { return JS_RUN_TEST_IDLE; }
        virtual JTAG_STATE endState() const { return JS_RUN_TEST_IDLE; }
        std::string name;
    };

    class SvfOpComment: public SvfOp
    {
    public:
        explicit SvfOpComment(const std::string &s) : comment(s) { name = "comment"; }
        virtual std::string getComment() const { return comment; }
    protected:
        std::string comment;
    };

    class SvfShiftOptions
    {
    public:
        SvfShiftOptions(const JtagBits &_tdi, const JtagBits &_smask, const JtagBits &_tdo, const JtagBits &_mask, const JTAG_STATE &_shiftState,  const JTAG_STATE &_endState = JS_RUN_TEST_IDLE)
            : tdi(_tdi), smask(_smask), tdo(_tdo), mask(_mask), shiftState(_shiftState), endState(_endState) { JtagShift(*this); }
        //  std::ostream& operator<< (std::ostream & out);
        //  JtagShift getSeq() const { return JtagShift() ; }

        JtagBits tdi, smask, tdo, mask;
        JTAG_STATE shiftState;
        JTAG_STATE endState;
    };

    class SvfOpSxr: public SvfOp {
    public:
        explicit SvfOpSxr(const SvfShiftOptions &_opt)
            : opt(_opt) { name = "Shift xR"; }
        //  virtual void print() const { std::cout << "Op: " << name << ", tdi:" << opt.tdi.size() << ", tdo: " << opt.tdo.size() << std::endl; }
        virtual JtagShift getSequence() const { return JtagShift(opt) ; }
        virtual JTAG_STATE endState() const { return opt.endState; }
        SvfShiftOptions opt;
    };

    class SvfOpSdr: public SvfOpSxr {
    public:
        explicit SvfOpSdr(const SvfShiftOptions &_opt)
            : SvfOpSxr(_opt) { name = "Shift DR"; }
        virtual JTAG_STATE beginState() const { return JS_SHIFT_DR; }
    };
    class SvfOpSir: public SvfOpSxr {
    public:
        explicit SvfOpSir(const SvfShiftOptions &_opt)
            : SvfOpSxr(_opt) { name = "Shift IR"; }
        virtual JTAG_STATE beginState() const { return JS_SHIFT_IR; }
    };

    struct SvfOpState : SvfOp {
        explicit SvfOpState(JTAG_STATE s) : state(s) { name = std::string("State change to ")+jtagStateStr(s); }
        virtual JTAG_STATE beginState() const { return state; }
        virtual JTAG_STATE endState() const { return state; }
        virtual JtagShift getSequence() const;
        JTAG_STATE state;
    };

    struct SvfOpRunTest : SvfOp {
        explicit SvfOpRunTest(double _delay) : delay(_delay) { name = "RunTest"; }
        virtual JTAG_STATE beginState() const { return JS_RUN_TEST_IDLE; }
        virtual JTAG_STATE endState() const { return JS_RUN_TEST_IDLE; }
        virtual double getDelay() const { return delay; }
        //  virtual JtagShift getSequence() const { JtagShift j; j.resize(count, JtagStep()); return j; }
        double delay;
    };

    struct SvfOpTrst : SvfOp { bool trst; };

    class SvfList: public std::vector<boost::shared_ptr<SvfOp> >
    {
    public:
    };

    class TokenList : public std::list<std::string> {
    public:
        bool contains(const std::string &name) const;
    };

    std::ostream& operator<< (std::ostream & os, const TokenList &r);

    class SvfTapState;

    class SvfParser {
    public:
        explicit SvfParser(const std::string &fileName);
        ~SvfParser();

    protected:
        void parseTokens(const std::vector<std::string> &tokens);

        void parseTokensEndXr(const std::string &name, TokenList &t);
        void parseTokensXxr(const std::string &name, TokenList &t);

        void checkCountGe(const TokenList &t, size_t count) const;
        void checkCountEq(const TokenList &t, size_t count) const;
        JTAG_STATE getSvfTapState(const std::string &) const;
        void createKnownTokenList();

        SvfList commands;
        std::map<std::string, std::map<std::string, JtagBits> > hxr, txr;
        std::map<std::string, JtagBits> save_smask, save_mask;
        JTAG_STATE endDr;
        JTAG_STATE endIr;
        double frequency;

        SvfTapState *svfTapState;
        TokenList endXrTokens;
        TokenList xxrTokens;

    };

};

#endif /* SVF_PARSER_H */
